@@ -1,21 +1,19 @@
 # -*- coding: utf-8 -*-
 from __future__ import print_function, absolute_import
 import re
+import warnings
 
 from tornado import gen
 
 from ._compat import range_type, text_type, PY2
 
-from .err import (
-    Warning, Error, InterfaceError, DataError,
-    DatabaseError, OperationalError, IntegrityError, InternalError,
-    NotSupportedError, ProgrammingError)
+from . import err
 
 
 #: Regular expression for :meth:`Cursor.executemany`.
 #: executemany only suports simple bulk insert.
 #: You can use it to load large dataset.
-RE_INSERT_VALUES = re.compile(r"""INSERT\s.+\sVALUES\s+(\(\s*%s\s*(,\s*%s\s*)*\))\s*\Z""",
+RE_INSERT_VALUES = re.compile(r"""(INSERT\s.+\sVALUES\s+)(\(\s*%s\s*(?:,\s*%s\s*)*\))(\s*(?:ON DUPLICATE.*)?)\Z""",
                               re.IGNORECASE | re.DOTALL)
 
 
@@ -42,12 +40,6 @@ class Cursor(object):
         self._result = None
         self._rows = None
 
-    def __del__(self):
-        '''
-        When this gets GC'd close it.
-        '''
-        self.close()
-
     @gen.coroutine
     def close(self):
         '''
@@ -64,12 +56,12 @@ class Cursor(object):
 
     def _get_db(self):
         if not self.connection:
-            raise ProgrammingError("Cursor closed")
+            raise err.ProgrammingError("Cursor closed")
         return self.connection
 
     def _check_executed(self):
         if not self._executed:
-            raise ProgrammingError("execute() first")
+            raise err.ProgrammingError("execute() first")
 
     def _conv_row(self, row):
         return row
@@ -81,18 +73,23 @@ class Cursor(object):
         """Does nothing, required by DB API."""
 
     @gen.coroutine
-    def nextset(self):
-        """Get the next query set"""
+    def _nextset(self, unbuffered=False):
         conn = self._get_db()
         current_result = self._result
         if current_result is None or current_result is not conn._result:
             raise gen.Return()
         if not current_result.has_next:
             raise gen.Return()
-        yield conn.next_result()
-        self._do_get_result()
+        yield conn.next_result(unbuffered=unbuffered)
+        yield self._do_get_result()
         raise gen.Return(True)
 
+    @gen.coroutine
+    def nextset(self):
+        """Get the next query set"""
+        res = yield self._nextset(False)
+        raise gen.Return(res)
+
     def _escape_args(self, args, conn):
         if isinstance(args, (tuple, list)):
             return tuple(conn.escape(arg) for arg in args)
@@ -148,10 +145,11 @@ class Cursor(object):
 
         m = RE_INSERT_VALUES.match(query)
         if m:
-            q_values = m.group(1).rstrip()
+            q_prefix = m.group(1)
+            q_values = m.group(2).rstrip()
+            q_postfix = m.group(3) or ''
             assert q_values[0] == '(' and q_values[-1] == ')'
-            q_prefix = query[:m.start(1)]
-            yield self._do_execute_many(q_prefix, q_values, args,
+            yield self._do_execute_many(q_prefix, q_values, q_postfix, args,
                                         self.max_stmt_length,
                                         self._get_db().encoding)
         else:
@@ -163,11 +161,13 @@ class Cursor(object):
         raise gen.Return(self.rowcount)
 
     @gen.coroutine
-    def _do_execute_many(self, prefix, values, args, max_stmt_length, encoding):
+    def _do_execute_many(self, prefix, values, postfix, args, max_stmt_length, encoding):
         conn = self._get_db()
         escape = self._escape_args
         if isinstance(prefix, text_type):
             prefix = prefix.encode(encoding)
+        if isinstance(postfix, text_type):
+            postfix = postfix.encode(encoding)
         sql = bytearray(prefix)
         args = iter(args)
         v = values % escape(next(args), conn)
@@ -180,14 +180,13 @@ class Cursor(object):
             if isinstance(v, text_type):
                 v = v.encode(encoding)
             if len(sql) + len(v) + 1 > max_stmt_length:
-                print(sql)
-                yield self.execute(bytes(sql))
+                yield self.execute(bytes(sql + postfix))
                 rows += self.rowcount
                 sql = bytearray(prefix)
             else:
                 sql += b','
             sql += v
-        yield self.execute(bytes(sql))
+        yield self.execute(bytes(sql + postfix))
         rows += self.rowcount
         self.rowcount = rows
 
@@ -246,7 +245,7 @@ class Cursor(object):
         ''' Fetch several rows '''
         self._check_executed()
         if self._rows is None:
-            return None
+            return ()
         end = self.rownumber + (size or self.arraysize)
         result = self._rows[self.rownumber:end]
         self.rownumber = min(end, len(self._rows))
@@ -256,7 +255,7 @@ class Cursor(object):
         ''' Fetch all the rows '''
         self._check_executed()
         if self._rows is None:
-            return None
+            return ()
         if self.rownumber:
             result = self._rows[self.rownumber:]
         else:
@@ -271,7 +270,7 @@ class Cursor(object):
         elif mode == 'absolute':
             r = value
         else:
-            raise ProgrammingError("unknown scroll mode %s" % mode)
+            raise err.ProgrammingError("unknown scroll mode %s" % mode)
 
         if not (0 <= r < len(self._rows)):
             raise IndexError("out of range")
@@ -282,8 +281,9 @@ class Cursor(object):
         conn = self._get_db()
         self._last_executed = q
         yield conn.query(q)
-        self._do_get_result()
+        yield self._do_get_result()
 
+    @gen.coroutine
     def _do_get_result(self):
         conn = self._get_db()
 
@@ -295,27 +295,37 @@ class Cursor(object):
         self.lastrowid = result.insert_id
         self._rows = result.rows
 
+        if result.warning_count > 0:
+            yield self._show_warnings(conn)
+
+    @gen.coroutine
+    def _show_warnings(self, conn):
+        ws = yield conn.show_warnings()
+        for w in ws:
+            warnings.warn(w[-1], err.Warning, 4)
+
     def __iter__(self):
         return iter(self.fetchone, None)
 
-    Warning = Warning
-    Error = Error
-    InterfaceError = InterfaceError
-    DatabaseError = DatabaseError
-    DataError = DataError
-    OperationalError = OperationalError
-    IntegrityError = IntegrityError
-    InternalError = InternalError
-    ProgrammingError = ProgrammingError
-    NotSupportedError = NotSupportedError
+    Warning = err.Warning
+    Error = err.Error
+    InterfaceError = err.InterfaceError
+    DatabaseError = err.DatabaseError
+    DataError = err.DataError
+    OperationalError = err.OperationalError
+    IntegrityError = err.IntegrityError
+    InternalError = err.InternalError
+    ProgrammingError = err.ProgrammingError
+    NotSupportedError = err.NotSupportedError
 
 
 class DictCursorMixin(object):
     # You can override this to use OrderedDict or other dict-like types.
     dict_type = dict
 
+    @gen.coroutine
     def _do_get_result(self):
-        super(DictCursorMixin, self)._do_get_result()
+        yield super(DictCursorMixin, self)._do_get_result()
         fields = []
         if self.description:
             for f in self._result.fields:
@@ -377,9 +387,14 @@ class SSCursor(Cursor):
         conn = self._get_db()
         self._last_executed = q
         yield conn.query(q, unbuffered=True)
-        self._do_get_result()
+        yield self._do_get_result()
         raise gen.Return(self.rowcount)
 
+    @gen.coroutine
+    def nextset(self):
+        res = yield self._nextset(True)
+        raise gen.Return(res)
+
     @gen.coroutine
     def read_next(self):
         """ Read next row """
@@ -393,7 +408,7 @@ class SSCursor(Cursor):
         self._check_executed()
         row = yield self.read_next()
         if row is None:
-            raise gen.Return()
+            raise gen.Return(None)
         self.rownumber += 1
         raise gen.Return(row)
 
@@ -401,7 +416,8 @@ class SSCursor(Cursor):
     def fetchall(self):
         """
         Fetch all, as per MySQLdb. Pretty useless for large queries, as
-        it is buffered.
+        it is buffered. See fetchall_unbuffered(), if you want an unbuffered
+        generator version of this method.
         """
         rows = []
         while True:
@@ -413,7 +429,7 @@ class SSCursor(Cursor):
 
     @gen.coroutine
     def fetchmany(self, size=None):
-        """Fetch many"""
+        """ Fetch many """
         self._check_executed()
         if size is None:
             size = self.arraysize
@@ -427,28 +443,29 @@ class SSCursor(Cursor):
             self.rownumber += 1
         raise gen.Return(rows)
 
+    @gen.coroutine
     def scroll(self, value, mode='relative'):
         self._check_executed()
 
         if mode == 'relative':
             if value < 0:
-                raise NotSupportedError(
+                raise err.NotSupportedError(
                         "Backwards scrolling not supported by this cursor")
 
             for _ in range_type(value):
-                self.read_next()
+                yield self.read_next()
             self.rownumber += value
         elif mode == 'absolute':
             if value < self.rownumber:
-                raise NotSupportedError(
+                raise err.NotSupportedError(
                     "Backwards scrolling not supported by this cursor")
 
             end = value - self.rownumber
             for _ in range_type(end):
-                self.read_next()
+                yield self.read_next()
             self.rownumber = value
         else:
-            raise ProgrammingError("unknown scroll mode %s" % mode)
+            raise err.ProgrammingError("unknown scroll mode %s" % mode)
 
 
 class SSDictCursor(DictCursorMixin, SSCursor):
